/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */
#include "defs.h"
#include "cpu.h"
#include "io.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(STATFS_USE_STATVFS)
#include <sys/statvfs.h>
#elif defined(STATFS_USE_VFS)
#include <sys/vfs.h>
#elif defined(STATFS_USE_MOUNT)
#include <sys/param.h>
#include <sys/mount.h>
#elif defined(STATFS_USE_STATFS_VMOUNT)
#include <sys/statfs.h>
#include <sys/vmount.h>
#endif
#include <grp.h>
#include <fcntl.h>
#include <setjmp.h>
#include "gemdos.h"
#include <dirent.h>
#include "debug.h"
#include "main.h"
#include "utils.h"
#include "io.h"
#include "tosdefs.h"
#include "cartridge.h"

#define DEBUG_GEMDOS 1
#ifndef TRACE_GEMDOS
#define TRACE_GEMDOS 1
#endif

#undef T
#if DEBUG_GEMDOS
#define T(_x) debug _x
#else
#define T(_x)
#endif

#if TRACE_GEMDOS
#define G(_x) debug _x
#else
#define G(_x)
#endif
#define GRESULT G(("= %ld\n", DREG(0)))
#define GOTOS do { G(("-> TOS\n")); return CALL_GEMDOS; } while(0)
#define NYI do { G(("(NYI) -> TOS\n")); return CALL_GEMDOS; } while(0)

#if 1
#define STATFUNC stat
#else
#define STATFUNC lstat
#endif

#define DEFAULT_MODE 0644
#define DEFAULT_DIRMODE 0755


/* some of this code is was stolen&modified from JET */
#define TOS_NAMELEN 12

typedef struct
{
	char index[2];						/* FIXME - should use seekdir() instead of DIR * array */
	char magic[4];
#define SVALID 0x1234fedc
#define EVALID 0x5678ba90
	char dta_pat[TOS_NAMELEN + 1];
	char dta_sattrib;
	char dta_drive;
	char dta_attrib;
	char dta_time[2];
	char dta_date[2];
	char dta_size[4];
	char dta_name[TOS_NAMELEN + 2];
} DTA;

typedef struct
{
	UL bp;
	int f;
} FINFO;

#define MAXFILES 100
#define H_OFFSET 100

#define MAXDIRS 40

#define MIN_VALID_HANDLE (-3)
#define MIN_FILE_HANDLE   6
#define INVALID_HANDLE (MIN_VALID_HANDLE-1)

#define STD_H_PRN -3
#define STD_H_AUX -2
#define STD_H_CON -1

#define DOM_TOS  0
#define DOM_MINT 1
#define TOS_PATH_MAX 128

#define DP_FT_DIR  0x00100000l
#define DP_FT_CHR  0x00200000l
#define DP_FT_BLK  0x00400000l
#define DP_FT_REG  0x00800000l
#define DP_FT_LNK  0x01000000l
#define DP_FT_SOCK 0x02000000l
#define DP_FT_FIFO 0x04000000l
#define DP_FT_MEM  0x08000000l

#define DP_INDEX   0x0001
#define DP_DEV     0x0002
#define DP_RDEV    0x0004
#define DP_NLINK   0x0008
#define DP_UID     0x0010
#define DP_GID     0x0020
#define DP_BLKSIZE 0x0040
#define DP_SIZE    0x0080
#define DP_NBLOCKS 0x0100
#define DP_ATIME   0x0200
#define DP_CTIME   0x0400
#define DP_MTIME   0x0800

typedef struct {
	int domain;
	UL usrval;
	int stdhandle[MIN_FILE_HANDLE];
	int fhandle[MAXFILES];
} PRG;

#define GEMDOS_OK        0
#define GEMDOS_EDRVNR   -2
#define GEMDOS_EWRITF  -10
#define GEMDOS_EREADF  -11
#define GEMDOS_EWRPRO  -13
#define GEMDOS_EUNDEV  -15
#define GEMDOS_EINVFN  -32
#define GEMDOS_EFILNF  -33
#define GEMDOS_EPTHNF  -34
#define GEMDOS_ENHNDL  -35
#define GEMDOS_EACCDN  -36
#define GEMDOS_EIHNDL  -37
#define GEMDOS_ENSMEM  -39
#define GEMDOS_EIMBA   -40
#define GEMDOS_EDRIVE  -46
#define GEMDOS_ENSAME  -48
#define GEMDOS_ENMFIL  -49
#define GEMDOS_ELOCKED -58
#define GEMDOS_ERANGE  -64
#define GEMDOS_EINTRN  -65
#define GEMDOS_EPLFMT  -66

#define FA_RDONLY  0x01
#define FA_HIDDEN  0x02
#define FA_SYSTEM  0x04
#define	FA_LABEL   0x08
#define FA_DIREC   0x10
#define FA_ARCHIVE 0x20
#define FA_SYMLINK 0x40

extern int redirect_cconws;
int gemdos_ok = 0;
static UL gemdos_drives = 0;
static UW curdrv = 0;
static char curpath[MAXDRIVES][PATH_MAX];

static DIR *fsdir[MAXDIRS];
static char *fspath[MAXDIRS];
static char curspath[MAXDIRS][PATH_MAX];
static char *root[MAXDRIVES];

#define CALL_GEMDOS FALSE
#define GEMDOS_DONE TRUE

#define PMAGIC 0xbeef4afc

int add_gemdos_drive(char d, char *path)
{
	int drv = toupper(d & 0xff) - 'A';

	if (drv < 0 || drv > MAXDRIVES)
	{
		fprintf(stderr, "Invalid drive: `%c'\n", d);
		return 0;
	} else if (drive_bits & (1 << drv))
	{
		fprintf(stderr, "Drive `%c' is already used as a BIOS drive\n", d);
		return 0;
	} else if (gemdos_drives & (1 << drv))
	{
		fprintf(stderr, "Drive `%c' already specified as a GEMDOS drive\n", d);
		return 0;
	}
	gemdos_drives |= 1 << drv;
	drive_bits |= 1 << drv;
	drive_fd[drv] = -1;
	root[drv] = strdup(path);
	return 1;
}

/**************************************************************/

UL get_bp(void)
{
	UL os_beg;
	void *act_pd;

	os_beg = OS_BEG();
	if (TOS_VERSION(os_beg) < 0x102)
	{
		if ((LM_UW(MEM(os_beg + 0x1c)) >> 1) == 4)
			act_pd = MEM(0x873c);
		else
			act_pd = MEM(0x602c);
	} else
	{
		act_pd = MEM(LM_UL(MEM(os_beg + 0x28)));
	}
	return LM_UL(act_pd);
}

#define PRG_OFF 80

static PRG *get_prg_bp(UL bp)
{
	if (LM_UL(MEM(bp+PRG_OFF)) == PMAGIC)
		return *((PRG **)MEM(bp+PRG_OFF+4));
	return NULL;
}

static PRG *get_prg(void)
{
	return get_prg_bp(get_bp());
}

static void set_prg(UL bp, PRG *prg)
{
	SM_UL(MEM(bp+PRG_OFF), PMAGIC);
	*((PRG **)MEM(bp+PRG_OFF+4)) = prg;
}

static void str_lwr(char *path)
{
	while (*path)
	{
		*path = tolower(*path);
		path++;
	}
}

static void fname2unix(char *stpath, char *unixpath)
{
	char *n;
#if DEBUG_GEMDOS
	char *b = unixpath;
#endif

	if (stpath[1] == ':')
		n = stpath + 2;
	else
		n = stpath;
	while (*n)
	{
		*unixpath++ = (*n == '\\' ? '/' : tolower(*n));
		n++;
	}
	*unixpath = 0;
	T(("  Converted %s to %s\n", stpath, unixpath));
}

static int st2unixpath(char *stpath, char *fullpath, char *unixpath)
{
	int drv;
	char path[PATH_MAX];
	char lastpath[PATH_MAX];
	
	drv = (stpath[1] == ':' ? toupper(stpath[0]) - 'A' : curdrv);
	if (((gemdos_drives >> drv) & 1) == 0)
	{
		T(("ERROR in pathname conversion (wrong drive)\n"));
		*unixpath = 0;
		return FALSE;
	}
	if (fullpath == NULL)
		fullpath = lastpath;
	if (stpath[1] == ':')
	{
		strcpy(fullpath, stpath);
	} else if (stpath[0] == '\\')
	{
		fullpath[0] = curdrv + 'A';
		fullpath[1] = ':';
		strcpy(fullpath+2, stpath);
	} else
	{
		fullpath[0] = curdrv + 'A';
		fullpath[1] = ':';
		strcat(strcat(strcpy(fullpath+2, curpath[drv]), "\\"), stpath);
	}
	fname2unix(fullpath, path);
	strcat(strcpy(unixpath, root[drv]), path);
	T(("  st2unixpath(%s,%s)\n", stpath, unixpath));
	return TRUE;
}

static int unix2fname(char *unixname, char *stname)
{
	/* FIXME */
	char *p = unixname;

#if DEBUG_GEMDOS
	char *q = stname;
#endif
	if (strlen(unixname) > TOS_NAMELEN)
		return 0;
	while (*p)
	{
		if (*p == '.')
		{
#if 0
			if (dot)
				return 0;
			else
				dot = 1;
#endif
			*stname++ = '.';
		} else
			*stname++ = toupper(*p);
		p++;
	}
	*stname = 0;
	T(("  Unix name: %s, ST name: %s\n", unixname, q));
	return 1;
}

static int match(char *pat, char *name)
{
	T(("Match(%s,%s)\n", pat, name));
	if (strcmp(pat, "*.*") == 0)
		return 1;
	else if (strcasecmp(pat, name) == 0)
		return 1;
	else
	{
		char *p = pat;
		char *n = name;

		for (; *n;)
		{
			if (*p == '*')
			{
				while (*n && *n != '.')
					n++;
				p++;
			} else if (*p == '?' && *n)
			{
				n++;
				p++;
			} else if (toupper(*p) != toupper(*n))
			{
				return 0;
			} else
			{
				p++;
				n++;
			}
		}
		if (*p == 0)
		{
			T(("(%s) matched (%s)\n", name, pat));
			return 1;
		}
	}
	return 0;
}

static void compress_path(char *path)
{
	char *d, *p = path;

	T(("  Before path compression: %s\n", p));
	d = path;
	while (*p)
	{
		if (*p == '\\')
		{
			if (strncmp(p, "\\..\\", 4) == 0)
			{
				while (--d >= path && *d != '\\') ;
				if (d < path)
				{
					error("FATAL error in compress_path: \\..\\\n");
				}
				p += 3;
			} else if (strncmp(p, "\\.\\", 3) == 0)
			{
				p += 2;
			} else
			{
				while (*++p == '\\') ;
				*d++ = '\\';
			}
		} else
			*d++ = *p++;
	}
	*d = 0;
	T(("  Result of path compression: %s\n", path));
}

static void xclosedir(int i)
{
	if (i >= 0)
	{
		if (fsdir[i] != NULL)
			closedir(fsdir[i]);
		if (fspath[i] != NULL)
			free(fspath[i]);
		fsdir[i] = NULL;
		fspath[i] = NULL;
	}
}

static int xopendir(char *p)
{
	int i;
	DIR *d;

	for (i = 0; i < MAXDIRS && fsdir[i]; i++)
		;
	if (i == MAXDIRS)
	{
		error("FATAL error in xopendir - out of dirs\n");
	}
	fspath[i] = strdup(p);
	d = opendir(p);
	if (d == NULL)
	{
		T(("opendir(%s)  - no files\n", p));
		xclosedir(i);
		return -1;
	}
	fsdir[i] = d;
	return i;
}

static void xrewinddir(int i)
{
	if (i >= 0 && i < MAXDIRS && fsdir[i] != NULL)
	{
		rewinddir(fsdir[i]);
	}
}

static struct dirent *xreaddir(int i)
{
	if (i >= 0 && i < MAXDIRS && fsdir[i] != NULL)
		return readdir(fsdir[i]);
	return NULL;
}

static UW time2dos(time_t t)
{
	struct tm *x;

	x = localtime(&t);
	return (x->tm_sec >> 1) | (x->tm_min << 5) | (x->tm_hour << 11);
}

static UW date2dos(time_t t)
{
	struct tm *x;

	x = localtime(&t);
	return x->tm_mday | ((x->tm_mon + 1) << 5) | (MAX(x->tm_year - 80, 0) << 9);
}

static UB mode2dos(struct stat *s)
{
	UB attrib;
	int mode = s->st_mode & 0777;
	
#define in_group(id) (getgid() == id) /* FIXME */

	attrib = 0;
	if (S_ISDIR(s->st_mode))
		attrib |= FA_DIREC;
	if (getuid() == s->st_uid)
		mode >>= 6;
	else if (in_group(s->st_gid))
		mode >>= 3;
	if (!(mode & 002))
		attrib |= FA_RDONLY;
	if (s->st_mtime > s->st_ctime)
		attrib |= FA_ARCHIVE;
	return attrib;
}

static int st2unixmode(UW mode)
{
	switch (mode)
	{
	case 0:
		return O_RDONLY;
	case 1:
		return O_WRONLY;				/* kludge to avoid files being created */
	case 2:
		return O_RDWR;
	}
	return 0;
}

static int unix_to_tos_errno(int errno)
{
	switch (errno)
	{
		case EPERM: return GEMDOS_EACCDN;
		case ENOENT: return GEMDOS_EFILNF;
		case EIO: return GEMDOS_EREADF;
		case ENXIO: return GEMDOS_EDRVNR;
		case E2BIG: return GEMDOS_ENSMEM;
		case ENOEXEC: return GEMDOS_EPLFMT;
		case EBADF: return GEMDOS_EIHNDL;
		case ECHILD: return GEMDOS_EFILNF;
		case EAGAIN: return GEMDOS_ENSMEM;
		case ENOMEM: return GEMDOS_ENSMEM;
		case EACCES: return GEMDOS_EACCDN;
		case EFAULT: return GEMDOS_EIMBA;
		case ENOTBLK: return GEMDOS_EUNDEV;
		case EBUSY: return GEMDOS_ELOCKED;
		case EEXIST: return GEMDOS_EACCDN;
		case EXDEV: return GEMDOS_ENSAME;
		case ENODEV: return GEMDOS_EUNDEV;
		case ENOTDIR: return GEMDOS_EPTHNF;
		case EISDIR: return GEMDOS_EACCDN;
		case EINVAL: return GEMDOS_EINVFN;
		case ENFILE: return GEMDOS_ENHNDL;
		case EMFILE: return GEMDOS_ENHNDL;
		case ENOTTY: return GEMDOS_EUNDEV;
		case ETXTBSY: return GEMDOS_ELOCKED;
		case EFBIG: return GEMDOS_ERANGE;
		case ENOSPC: return GEMDOS_EWRITF;
		case ESPIPE: return GEMDOS_EREADF;
		case EROFS: return GEMDOS_EWRPRO;
		case EMLINK: return GEMDOS_EACCDN;
		case EPIPE: return GEMDOS_EWRITF;
		case EDOM: return GEMDOS_ERANGE;
		case ERANGE: return GEMDOS_ERANGE;
		case EDEADLK: return GEMDOS_ELOCKED;
		case ENAMETOOLONG: return GEMDOS_ERANGE;
		case ENOLCK: return GEMDOS_ENHNDL;
		case ENOSYS: return GEMDOS_EINVFN;
		case ENOTEMPTY: return GEMDOS_EACCDN;
		case ELOOP: return GEMDOS_EACCDN;
#if defined(EWOULDBLOCK) && EWOULDBLOCK != EAGAIN
		case EWOULDBLOCK: return GEMDOS_ELOCKED;
#endif
		default: return GEMDOS_EACCDN;
	}
}

/**************************************************************/

static int Cconin(void)
{
	G(("Cconin() "));
	GOTOS;
}

/**************************************************************/

static int Cconout(UW c)
{
	G(("Cconout($%x) ", c));
	GOTOS;
}

/**************************************************************/

static int Cauxin(void)
{
	G(("Cauxin() "));
	GOTOS;
}

/**************************************************************/

static int Cauxout(UW c)
{
	G(("Cauxout($%x) ", c));
	GOTOS;
}

/**************************************************************/

static int Cprnout(UW c)
{
	G(("Cprnout($%x) ", c));
	GOTOS;
}

/**************************************************************/

static int Crawio(UW c)
{
	G(("Crawio($%x) ", c));
	GOTOS;
}

/**************************************************************/

static int Crawcin(void)
{
	G(("Crawcin() "));
	GOTOS;
}

/**************************************************************/

static int Cnecin(void)
{
	G(("Cnecin() "));
	GOTOS;
}

/**************************************************************/

static int Cconws(char *str)
{
	G(("Cconws(%s) ", str));
	if (!redirect_cconws)
		GOTOS;
	fputs(str, stdout);
	DREG(0) = 0;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Cconrs(char *buf)
{
	G(("Cconrs($%lx) \n", (UL) buf));
	GOTOS;
}

/**************************************************************/

static int Cconis(void)
{
	PRG *prg;
	
	G(("Cconis() "));
	prg = get_prg();
	if (prg != NULL)
	{
		if (prg->stdhandle[0] >= H_OFFSET)
		{
			DREG(0) = -1;
		} else
		{
			GOTOS;
		}
	} else
	{
		DREG(0) = 0;
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Cauxis(void)
{
	PRG *prg;
	
	G(("Cauxis() "));
	prg = get_prg();
	if (prg != NULL)
	{
		if (prg->stdhandle[2] >= H_OFFSET)
		{
			DREG(0) = -1;
		} else
		{
			/* FIXME: check serial_stat() */
			GOTOS;
		}
	} else
	{
		DREG(0) = 0;
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Cconos(void)
{
	G(("Cconos() \n"));
	DREG(0) = 1;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Cprnos(void)
{
	PRG *prg;
	
	G(("Cprnos() \n"));
	prg = get_prg();
	if (prg != NULL)
	{
		if (prg->stdhandle[3] >= H_OFFSET)
		{
			DREG(0) = -1;
		} else if (parallel_stat())
		{
			DREG(0) = -1;
		} else
		{
			DREG(0) = 0;
		}
	} else
	{
		DREG(0) = 0;
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Cauxos(void)
{
	PRG *prg;
	
	G(("Cauxos() \n"));
	prg = get_prg();
	if (prg != NULL)
	{
		if (prg->stdhandle[2] >= H_OFFSET)
		{
			DREG(0) = -1;
		} else
		{
			GOTOS;
		}
	} else
	{
		DREG(0) = 0;
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/
/**************************************************************/

static int Dsetdrv(UW drv)
{
	G(("Dsetdrv(%c) ", 'A' + drv));
	curdrv = drv;
	if (BIT(drv, gemdos_drives) == 0)
	{
		GOTOS;
	}
	DREG(0) = gemdos_drives | drive_bits;
	G(("= $%lx\n", DREG(0)));
	return GEMDOS_DONE;
}

/**************************************************************/

static int Dgetdrv(void)
{
	DREG(0) = curdrv;
	G(("Dgetdrv() = %d\n", curdrv));
	return GEMDOS_DONE;
}

/**************************************************************/

static int Dfree(UL * buf, UW drv)
{
#ifdef STATFS_USE_STATVFS
	struct statvfs s;
#else
	struct statfs s;
#endif
	struct stat st;

	G(("Dfree(%d) ", drv));
	if (drv == 0)
		drv = curdrv;
	else
		--drv;
	if (!((gemdos_drives >> drv) & 1))
	{
		GOTOS;
	}
#ifdef STATFS_USE_STATVFS
	statvfs(root[drv], &s);
#else
	statfs(root[drv], &s);
#endif
#ifdef STATFS_USE_MOUNT					/* XXX 512b blocks */
	SM_UL(buf, s.f_bavail);
	SM_UL(buf + 1, s.f_blocks - (s.f_bfree - s.f_bavail));
	SM_UL(buf + 2, s.f_bsize);
	SM_UL(buf + 3, 1);
#else
	stat(root[drv], &st);
	SM_UL(buf, s.f_bavail);
	SM_UL(buf + 1, s.f_blocks);
	SM_UL(buf + 2, st.st_blksize / 2);
	SM_UL(buf + 3, 2);
#endif
	G(("= %ld/%ld/%ld/%ld\n", LM_UL(buf), LM_UL(buf+1), LM_UL(buf+2), LM_UL(buf+3)));
	DREG(0) = GEMDOS_OK;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Dcreate(char *pname)
{
	char upath[PATH_MAX];

	G(("Dcreate(%s) ", pname));
	if (!st2unixpath(pname, NULL, upath))
	{
		GOTOS;
	}
	T(("on drive %d -> %s ", curdrv, upath));
	DREG(0) = 0;
	if (mkdir(upath, DEFAULT_DIRMODE) < 0)
		DREG(0) = GEMDOS_EACCDN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Ddelete(char *pname)
{
	char upath[PATH_MAX];

	G(("Ddelete(%s) ", pname));
	if (!st2unixpath(pname, NULL, upath))
	{
		GOTOS;
	}
	T(("on drive %d -> %s ", curdrv, upath));
	DREG(0) = GEMDOS_OK;
	if (rmdir(upath) < 0)
		DREG(0) = GEMDOS_EACCDN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Dsetpath(char *pname)
{
	char upath[PATH_MAX];
	char fullpath[PATH_MAX];
	struct stat s;
	int drv;

	G(("Dsetpath(%s) ", pname));
	if (pname[1] == ':')
		drv = toupper(pname[0]) - 'A';
	else
		drv = curdrv;
	if (!st2unixpath(pname, fullpath, upath))
	{
		GOTOS;
	}
	if (STATFUNC(upath, &s) < 0 || !S_ISDIR(s.st_mode))
	{
		DREG(0) = GEMDOS_EPTHNF;
		GRESULT;
		return GEMDOS_DONE;
	}
	strcpy(curpath[drv], fullpath+2);
#if 1
	compress_path(curpath[drv]);
#endif
	T(("on drive %d (%s) ", curpath[drv], curdrv, upath));
	/* FIXME */
	DREG(0) = GEMDOS_OK;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int do_getpath(char *buf, UW drv, UW size, int do_lower)
{
	PRG *prg;
	
	if (drv == 0)
		drv = curdrv;
	else
		--drv;
	if (!((gemdos_drives >> drv) & 1))
	{
		GOTOS;
	}
	DREG(0) = 0;
	if (size == 0 || strlen(curpath[drv]) >= size)
	{
		DREG(0) = GEMDOS_ERANGE;
		GRESULT;
	} else
	{
		strcpy(buf, curpath[drv]);
		if (do_lower && (prg = get_prg()) != NULL && prg->domain == DOM_MINT)
			str_lwr(buf);
		G(("= %s\n", buf));
	}
	return GEMDOS_DONE;
}

static int Dgetpath(char *buf, UW drv)
{
	G(("Dgetpath(%d) ", drv));
	return do_getpath(buf, drv, TOS_PATH_MAX, TRUE);
}

/**************************************************************/

static int Dgetcwd(char *buf, UW drv, UW size)
{
	G(("Dgetcwd(%d,%d) ", drv, size));
	return do_getpath(buf, drv, size, FALSE);
}

/**************************************************************/

static int Dpathconf(char *pname, W mode)
{
	char upath[PATH_MAX];
	char fullpath[PATH_MAX];
	int drv;
	
	G(("Dpathconf(%s,%d) ", pname, mode));
	if (pname[1] == ':')
		drv = toupper(pname[0]) - 'A';
	else
		drv = curdrv;
	if (!st2unixpath(pname, fullpath, upath))
	{
		GOTOS;
	}
	switch (mode)
	{
	case -1:
		DREG(0) = 6;
		break;
	case 0:
		DREG(0) = MAXFILES;
		break;
	case 1:
		DREG(0) = LINK_MAX;
		break;
	case 2:
		DREG(0) = PATH_MAX;
		break;
	case 3:
		DREG(0) = NAME_MAX;
		break;
	case 4:
		DREG(0) = 512;
		break;
	case 5:
		DREG(0) = 1;
		break;
	case 6:
		DREG(0) = 0;
		break;
	case 7:
		DREG(0) = FA_RDONLY|FA_HIDDEN|FA_SYSTEM|FA_ARCHIVE|(0666<<8)|
			DP_FT_DIR|DP_FT_CHR|DP_FT_BLK|DP_FT_REG|DP_FT_LNK|DP_FT_SOCK|
			DP_FT_FIFO;
		break;
	case 8:
		DREG(0) = DP_INDEX|DP_DEV|DP_RDEV|DP_NLINK|DP_UID|DP_GID|DP_SIZE|
			DP_BLKSIZE|DP_NBLOCKS|DP_ATIME|DP_CTIME|DP_MTIME;
		break;
	default:
		DREG(0) = GEMDOS_EINVFN;
		break;
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Dopendir(char *pname, W mode)
{
	int handle;
	char upath[PATH_MAX];
	char fullpath[PATH_MAX];

	G(("Dopendir(%s,%d) ", pname, mode));
	if (!st2unixpath(pname, fullpath, upath))
	{
		GOTOS;
	}
	handle = xopendir(upath);
	if (handle < 0)
	{
		DREG(0) = unix_to_tos_errno(errno);
		GRESULT;
		return GEMDOS_DONE;
	}
#define DOPENDIR_NORMAL 0x8000
	if (mode == 0)
		handle |= DOPENDIR_NORMAL;
	DREG(0) = handle;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Dreaddir(W len, L handle, char *buf)
{
	int mode;
	struct dirent *ent;
	
	G(("Dreaddir(%d,%ld,$%lx) ", len, handle, VADDR(buf)));
	mode = handle & DOPENDIR_NORMAL ? 1 : 0;
	handle &= ~DOPENDIR_NORMAL;
	if (handle < 0 || handle >= MAXDIRS)
	{
		DREG(0) = GEMDOS_EIHNDL;
		GRESULT;
		return GEMDOS_DONE;
	}
	ent = xreaddir(handle);
	if (ent == 0)
	{
		DREG(0) = GEMDOS_ENMFIL;
		GRESULT;
		return GEMDOS_DONE;
	}
	if (strlen(ent->d_name) >= len)
	{
		DREG(0) = GEMDOS_ERANGE;
		GRESULT;
		return GEMDOS_DONE;
	}
	if (mode)
	{
		SM_UL((UL *)buf, ent->d_ino);
		buf += 4;
	}
	strcpy(buf, ent->d_name);
	G(("found %s ", buf));
	DREG(0) = GEMDOS_OK;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Drewinddir(L handle)
{
	G(("Drewinddir(%ld) ", handle));
	handle &= ~DOPENDIR_NORMAL;
	if (handle < 0 || handle >= MAXDIRS)
	{
		DREG(0) = GEMDOS_EIHNDL;
		GRESULT;
		return GEMDOS_DONE;
	}
	xrewinddir(handle);
	DREG(0) = GEMDOS_OK;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Dclosedir(L handle)
{
	G(("Dclosedir(%ld) ", handle));
	
	handle &= ~DOPENDIR_NORMAL;
	if (handle < 0 || handle >= MAXDIRS)
	{
		DREG(0) = GEMDOS_EIHNDL;
	} else
	{
		xclosedir(handle);
		DREG(0) = GEMDOS_OK;
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static void stat_to_xattr(char *xattr, struct stat *s)
{
	UW mode;
		
	mode = 0;
	if (S_ISSOCK(s->st_mode))
		;
	else if (S_ISLNK(s->st_mode))
		mode |= 0160000;
	else if (S_ISREG(s->st_mode))
		mode |= 0100000;
	else if (S_ISBLK(s->st_mode))
		mode |= 0060000;
	else if (S_ISDIR(s->st_mode))
		mode |= 0040000;
	else if (S_ISCHR(s->st_mode))
		mode |= 0020000;
	else if (S_ISFIFO(s->st_mode))
		mode |= 0120000;
	if (s->st_mode & S_ISUID)
		mode |= 04000;
	if (s->st_mode & S_ISGID)
		mode |= 02000;
	if (s->st_mode & S_ISVTX)
		mode |= 01000;
	mode |= s->st_mode & 0777;
	SM_UW(xattr, mode);
	SM_UL(xattr+2, s->st_ino);
	SM_UW(xattr+6, s->st_dev);
	SM_UW(xattr+8, s->st_rdev);
	SM_UW(xattr+10, s->st_nlink);
	SM_UW(xattr+12, s->st_uid);
	SM_UW(xattr+14, s->st_gid);
	SM_UL(xattr+16, s->st_size);
	SM_UL(xattr+20, s->st_blksize);
	SM_UL(xattr+24, s->st_blocks);
	SM_UW(xattr+28, time2dos(s->st_mtime));
	SM_UW(xattr+30, date2dos(s->st_mtime));
	SM_UW(xattr+32, time2dos(s->st_atime));
	SM_UW(xattr+34, date2dos(s->st_atime));
	SM_UW(xattr+36, time2dos(s->st_ctime));
	SM_UW(xattr+38, date2dos(s->st_ctime));
	SM_UW(xattr+40, mode2dos(s));
	SM_UW(xattr+42, 0);
	SM_UL(xattr+44, 0);
	SM_UL(xattr+48, 0);
}

static int Dxreaddir(W len, L handle, char *buf, UL xattr, UL xr)
{
	int mode;
	struct dirent *ent;
	
	G(("Dxreaddir(%d,%ld,$%lx,$%lx,$%lx) ", len, handle, VADDR(buf), xattr, xr));
	mode = handle & DOPENDIR_NORMAL ? 1 : 0;
	handle &= ~DOPENDIR_NORMAL;
	if (handle < 0 || handle >= MAXDIRS)
	{
		DREG(0) = GEMDOS_EIHNDL;
		GRESULT;
		return GEMDOS_DONE;
	}
	ent = xreaddir(handle);
	if (ent == 0)
	{
		DREG(0) = GEMDOS_ENMFIL;
		GRESULT;
		return GEMDOS_DONE;
	}
	if (strlen(ent->d_name) >= len)
	{
		DREG(0) = GEMDOS_ERANGE;
		GRESULT;
		return GEMDOS_DONE;
	}
	if (mode)
	{
		SM_UL((UL *)buf, ent->d_ino);
		buf += 4;
	}
	strcpy(buf, ent->d_name);
	if (xr)
		SM_L(MEM(xr), GEMDOS_OK);
	if (xattr != 0)
	{
		struct stat s;
		char path[PATH_MAX];
		
		strcat(strcat(strcpy(path, fspath[handle]), "/"), ent->d_name);
		if (lstat(path, &s))
		{
			if (xr)
				SM_L(MEM(xr), unix_to_tos_errno(errno));
		} else
		{
			stat_to_xattr(MEM(xattr), &s);
		}
	}
	G(("found %s ", buf));
	DREG(0) = GEMDOS_OK;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Dcntl(W cmd, char *name, UL arg)
{
	G(("Dcntl(%d,%s,%ld) ", cmd, name, arg));
	NYI;
}

/**************************************************************/

static int Dlock(W mode, W drv)
{
	G(("Dlock(%d,%d) ", mode, drv));
	NYI;
}

/**************************************************************/

static int Dreadlabel(char *path, char *buf, UW length)
{
	G(("Dreadlabel(%s,$%lx,%d) ", path, VADDR(buf), length));
	GOTOS;
}

/**************************************************************/

static int Dwritelabel(char *path, char *label)
{
	G(("Dwritelabel(%s,%s) ", path, label));
	GOTOS;
}

/**************************************************************/
/**************************************************************/

static int Fsetdta(UL dtaptr)
{
	UL bp = get_bp();

	G(("Fsetdta($%lx) bp=$%lx\n", dtaptr, bp));
#if 1
	SM_UL(MEM(bp + 32), dtaptr);
#endif
	return GEMDOS_DONE;
}

/**************************************************************/

static int Fgetdta(void)
{
	UL bp = get_bp();

	DREG(0) = LM_UL(MEM(bp + 32));
	G(("Fgetdta() bp=$%lx = $%lx\n", bp, DREG(0)));
	return GEMDOS_DONE;
}

/**************************************************************/

static int Fcreate(char *fname, UW attribs)
{
	int i;
	int f;
	char upath[PATH_MAX];
	PRG *prg;
	
	G(("Fcreate(%s,$%x) ", fname, attribs));
	if (!st2unixpath(fname, NULL, upath))
	{
		GOTOS;
	}
	T(("on drive %d -> %s ", curdrv, upath));
	if (attribs == FA_LABEL)
	{
		DREG(0) = GEMDOS_EFILNF;
		GRESULT;
		return GEMDOS_DONE;
	}
	prg = get_prg();
	if (prg == NULL)
	{
		DREG(0) = GEMDOS_EFILNF;
		GRESULT;
		return GEMDOS_DONE;
	}
	for (i = 0; i < MAXFILES && prg->fhandle[i] >= 0; i++)
		;
	if (i >= MAXFILES)
	{
		DREG(0) = GEMDOS_ENHNDL;
		GRESULT;
		return GEMDOS_DONE;
	}
	f = creat(upath, DEFAULT_MODE);
	if (f < 0)
	{
		T(("%s ", strerror(errno)));
		DREG(0) = GEMDOS_EACCDN;
	} else
	{
		prg->fhandle[i] = f;
		DREG(0) = i + H_OFFSET;
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Fopen(char *fname, UW mode)
{
	int i;
	int f;
	char upath[PATH_MAX];
	PRG *prg;
	
	G(("Fopen(%s,$%x) ", fname, mode));
	if (!st2unixpath(fname, NULL, upath))
	{
		GOTOS;
	}
	if (strcasecmp(fname, "con:") == 0)
		DREG(0) = STD_H_CON & 0xffff;
	else if (strcasecmp(fname, "aux:") == 0)
		DREG(0) = STD_H_AUX & 0xffff;
	else if (strcasecmp(fname, "prn:") == 0)
		DREG(0) = STD_H_PRN & 0xffff;
	else {
		prg = get_prg();
		if (prg == NULL)
		{
			DREG(0) = GEMDOS_EINTRN;
		} else
		{
			T(("on drive %d -> %s ", curdrv, upath));
			for (i = 0; i < MAXFILES && prg->fhandle[i] >= 0; i++)
				;
			if (i >= MAXFILES)
			{
				DREG(0) = GEMDOS_ENHNDL;
			} else
			{
				f = open(upath, st2unixmode(mode));
				if (f < 0)
				{
					T(("%s ", strerror(errno)));
					DREG(0) = GEMDOS_EFILNF;
				} else
				{
					prg->fhandle[i] = f;
					DREG(0) = i + H_OFFSET;
				}
			}
		}
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static void check_close(PRG *prg, UL bp, int handle)
{
	int i;
	
	for (i = 0; i < MIN_FILE_HANDLE; i++)
		if (prg->stdhandle[i] == handle)
		{
			switch (i)
			{
				case 0: prg->stdhandle[i] = STD_H_CON; break;
				case 1: prg->stdhandle[i] = STD_H_CON; break;
				case 2: prg->stdhandle[i] = STD_H_AUX; break;
				case 3: prg->stdhandle[i] = STD_H_PRN; break;
			}
			SM_B(MEM(bp+48+i), prg->stdhandle[i]);
		}
}

static int Fclose(W handle)
{
	int i = handle - H_OFFSET;
	UL bp;
	PRG *prg;
	int f = -1;
	
	G(("Fclose(%d) ", handle));
	bp = get_bp();
	prg = get_prg_bp(bp);
	if (prg == NULL)
	{
		DREG(0) = GEMDOS_EINTRN;
		return GEMDOS_DONE;
	}
	if (handle >= 0 && handle < MIN_FILE_HANDLE)
	{
		if (prg->stdhandle[handle] >= H_OFFSET)
		{
			i = prg->stdhandle[handle] - H_OFFSET;
			f = prg->fhandle[i];
		}
	} else if (i >= 0 && i < MAXFILES)
	{
		f = prg->fhandle[i];
	}
	if (f < 0)
	{
		GOTOS;
	} else
	{
		close(f);
		prg->fhandle[i] = INVALID_HANDLE;
		check_close(prg, bp, handle);
		DREG(0) = 0;
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

/* Problem: if we pass a large `count', some systems give "Bad Address"
 * errors, because buf+count would be outside the process' address space (?)
 */
static int Fread(W handle, UL count, UL buf)
{
#define FREAD_CHUNK 4096
	int r;
	int i = handle - H_OFFSET;
	int total;
	int rd;
	int f = -1;
	PRG *prg;
	
	G(("Fread(%d,%ld,$%lx) ", handle, count, buf));
	prg = get_prg();
	if (prg == NULL)
	{
		DREG(0) = GEMDOS_EINTRN;
		GRESULT;
		return GEMDOS_DONE;
	}
	if (handle >= 0 && handle < MIN_FILE_HANDLE)
	{
		if (prg->stdhandle[handle] >= H_OFFSET)
		{
			i = prg->stdhandle[handle] - H_OFFSET;
			f = prg->fhandle[i];
		}
	} else if (i >= 0 && i < MAXFILES)
	{
		f = prg->fhandle[i];
	}
	if (f < 0)
	{
		GOTOS;
	}
	if (count > 0x1fffffff)
		count = 0x1fffffff;
	rd = FREAD_CHUNK;
	total = 0;
	for (; count > 0;)
	{
		if (count < rd)
			rd = count;
		r = read(f, MEM(buf + total), rd);
		if (r < 0)
		{
			T(("%s ", strerror(errno)));
			DREG(0) = -1;
			GRESULT;
			return GEMDOS_DONE;
		}
		count -= r;
		total += r;
		if (r < rd)
			break;
	}
	DREG(0) = total;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Fwrite(W handle, UL count, UL buf)
{
	int r;
	int i = handle - H_OFFSET;
	int f = -1;
	PRG *prg;
	
	G(("Fwrite(%d,%ld,$%lx) ", handle, count, buf));
	prg = get_prg();
	if (prg == NULL)
	{
		GOTOS;
	}
	if (handle >= 0 && handle < MIN_FILE_HANDLE)
	{
		if (prg->stdhandle[handle] >= H_OFFSET)
		{
			i = prg->stdhandle[handle] - H_OFFSET;
			f = prg->fhandle[i];
		}
	} else if (i >= 0 && i < MAXFILES)
	{
		f = prg->fhandle[i];
	}
	if (f < 0)
	{
		GOTOS;
	} else
	{
		r = write(f, MEM(buf), (size_t) count);
		DREG(0) = r;
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Fdelete(char *fname)
{
	char upath[PATH_MAX];

	G(("Fdelete(%s)", fname));
	if (!st2unixpath(fname, NULL, upath))
	{
		GOTOS;
	}
	T(("on drive %d -> %s ", curdrv, upath));
	DREG(0) = 0;
	if (unlink(upath) < 0)
		DREG(0) = GEMDOS_EACCDN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Fseek(UL offset, W handle, UW mode)
{
	int i = handle - H_OFFSET;
	int f = -1;
	PRG *prg;
	
	G(("Fseek(%ld,%d,%d) ", offset, handle, mode));
	prg = get_prg();
	if (prg == NULL)
	{
		GOTOS;
	}
	if (handle >= 0 && handle < MIN_FILE_HANDLE)
	{
		if (prg->stdhandle[handle] >= H_OFFSET)
		{
			i = prg->stdhandle[handle] - H_OFFSET;
			f = prg->fhandle[i];
		}
	} else if (i >= 0 && i < MAXFILES)
	{
		f = prg->fhandle[i];
	}
	if (f < 0)
	{
		GOTOS;
	} else
	{
		DREG(0) = lseek(f, offset, mode);
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Fattrib(char *name, UW mode, UW attrib)
{
	char u[PATH_MAX];
	struct stat s;

	G(("Fattrib(%s,%d,$%x) ", name, mode, attrib));
	if (!st2unixpath(name, NULL, u))
	{
		GOTOS;
	}
	DREG(0) = 0;
	if (mode == 0)
	{
		if (STATFUNC(u, &s) < 0)
		{
			DREG(0) = GEMDOS_EFILNF;
		} else
		{
			DREG(0) = (s.st_mode & S_IFDIR) ? FA_DIREC : 0;		/* FIXME */
		}
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Fdup(W fd)
{
	G(("Fdup(%d) ", fd));
	/*TODO*/
	GOTOS;
}

/**************************************************************/

static int Fforce(W stdh, W nonstdh)
{
	int i;
	
	G(("Fforce(%d,%d) ", stdh, nonstdh));
	if (stdh >= 0 && stdh < MIN_FILE_HANDLE)
	{
		UL bp = get_bp();
		PRG *prg = get_prg_bp(bp);
		if (prg != NULL)
		{
			if (prg->stdhandle[stdh] >= H_OFFSET)
				check_close(prg, bp, prg->stdhandle[stdh]);
			if ((nonstdh >= MIN_VALID_HANDLE && nonstdh < 0) ||
				(nonstdh >= H_OFFSET && nonstdh < (H_OFFSET+MAXFILES)))
			{
				i = nonstdh - H_OFFSET;
				if (i >= 0 && prg->fhandle[i] < 0)
				{
					DREG(0) = GEMDOS_EIHNDL;
				} else
				{
					prg->stdhandle[stdh] = nonstdh;
					SM_B(MEM(bp+48+stdh), nonstdh);
					DREG(0) = 0;
				}
			} else
			{
				GOTOS;
			}
		} else
		{
			DREG(0) = GEMDOS_EINTRN;
		}
		GRESULT;
		return GEMDOS_DONE;
	}
	GOTOS;
}

/**************************************************************/

static INLINE int attrib_matches(UB file_attrib, UB search_attrib)
{
	return (file_attrib == 0 ||
		((file_attrib != 0) && (search_attrib != FA_LABEL)) || ((search_attrib | FA_ARCHIVE|FA_RDONLY) & file_attrib));
}

static int do_stat(char *uname, DTA *dta)
{
	struct stat s;
	
	if (STATFUNC(uname, &s) < 0)
		return FALSE;
	dta->dta_attrib = mode2dos(&s);
	SM_UW((UW *) & (dta->dta_time), time2dos(s.st_mtime));
	SM_UW((UW *) & (dta->dta_date), date2dos(s.st_mtime));
	SM_UL((UL *) & (dta->dta_size), 0);
	if (!(LM_UB(&(dta->dta_attrib)) & FA_DIREC))
	{
		SM_UL((UL *) & (dta->dta_size), s.st_size);
	}
	return TRUE;
}

static void do_fsnext(DTA *dta, char *upath)
{
	struct dirent *e;
	char stname[TOS_NAMELEN + 1];
	UB attribs;
	int rr = -1;

	attribs = LM_UB(&(dta->dta_sattrib));
	T(("  Fsnext(%s) ", dta->dta_pat));
	rr = (LM_W((W *) & (dta->index)));
	DREG(0) = 0;
	do
	{
		do
		{
			e = xreaddir(rr);
			if (e == NULL)
			{
				G(("(no more files) "));
				xclosedir(LM_UW((UW *) & (dta->index)));
				DREG(0) = GEMDOS_ENMFIL;
				return;
			}
		} while (!unix2fname(e->d_name, stname) || !match(dta->dta_pat, stname));
		strcpy(dta->dta_name, stname);
		strcat(strcat(strcpy(upath, curspath[rr]), "/"), e->d_name);
		T(("stat(%s) ", upath));
	} while (!do_stat(upath, dta) || !attrib_matches(dta->dta_attrib, attribs));
	if (DREG(0) == 0)
	{
		G((" found %s ", dta->dta_name));
	}
}

static int Fsnext(void)
{
	char upath[PATH_MAX];
	int cursdrv;
	DTA *dta;
	
	dta = (DTA *) MEM(LM_UL(MEM(get_bp() + 32)));
	G(("Fsnext() "));
	if (LM_L((L *)dta->magic) != SVALID)
	{
		GOTOS;
	}
	cursdrv = dta->dta_drive;
	if (((gemdos_drives >> cursdrv) & 1) == 0)
	{
		GOTOS;
	}
	do_fsnext(dta, upath);
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int do_fsfirst(char *fspec, UW attribs, DTA *dta, char *upath)
{
	int r;
	char *p;
	char fullpath[PATH_MAX];
	int cursdrv;
	
	if (!st2unixpath(fspec, fullpath, upath))
	{
		SM_L((L *)dta->magic, EVALID);
		GOTOS;
	}
	T(("on drive %d ", curdrv));
	if (fspec[1] == ':')
		cursdrv = toupper(fspec[0]) - 'A';
	else
		cursdrv = curdrv;
	SM_L((L *)dta->magic, SVALID);
	dta->dta_drive = cursdrv;
	p = strrchr(fullpath, '\\') + 1;
	strncpy(dta->dta_pat, p, TOS_NAMELEN);
	dta->dta_pat[TOS_NAMELEN] = '\0';
	
	*p = 0;
	SM_UB(&(dta->dta_sattrib), attribs);
	if (attribs == FA_LABEL)
	{
		DREG(0) = GEMDOS_EFILNF;
		return GEMDOS_DONE;
	}
	st2unixpath(fullpath, fullpath, upath);
#if 1
	/* kill trailing /es */
	while (upath && *upath && *(p = upath + strlen(upath) - 1) == '/')
		*p = '\0';
#endif
	T(("xopendir(%s) ", upath));
	r = xopendir(upath);
	if (r < 0)
	{
		DREG(0) = GEMDOS_EFILNF;
		return GEMDOS_DONE;
	}
	SM_UW((UW *) & (dta->index), r);
	strcpy(curspath[r], upath);
	do_fsnext(dta, upath);
	if (!strchr(dta->dta_pat, '?') &&
		!strchr(dta->dta_pat, '*'))
	{
		xclosedir(r);
	}
	if (DREG(0) == GEMDOS_ENMFIL)
		DREG(0) = GEMDOS_EFILNF;
	return GEMDOS_DONE;
}

static int Fsfirst(char *fspec, UW attribs)
{
	int result;
	char upath[PATH_MAX];
	
	G(("Fsfirst(%s,$%x) ", fspec, attribs));
	/* char tpath[PATH_MAX]; */
	result = do_fsfirst(fspec, attribs, (DTA *) MEM(LM_UL(MEM(get_bp() + 32))), upath);
	if (result != CALL_GEMDOS)
		GRESULT;
	return result;
}

/**************************************************************/

static int Frename(char *old, char *nw)
{
	char o[PATH_MAX], n[PATH_MAX];
	int r;

	G(("Frename(%s) to %s ", old, nw));
	if (!st2unixpath(old, NULL, o) || !st2unixpath(nw, NULL, n))
	{
		GOTOS;
	}
	r = rename(o, n);
	DREG(0) = 0;
	if (r < 0)
	{
		DREG(0) = unix_to_tos_errno(errno);
		T(("%s ", strerror(errno)));
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Fdatime(UW * t, UW * d, W handle, UW wflag)
{
	struct stat s;
	int i = handle - H_OFFSET;
	PRG *prg;
	int f = -1;
	
	G(("Fdatime(%d,%d) ", handle, wflag));
	prg = get_prg();
	if (prg == NULL)
	{
		DREG(0) = GEMDOS_EINTRN;
		GRESULT;
		return GEMDOS_DONE;
	}
	if (handle >= 0 && handle < MIN_FILE_HANDLE)
	{
		if (prg->stdhandle[handle] >= H_OFFSET)
		{
			i = prg->stdhandle[handle] - H_OFFSET;
			f = prg->fhandle[i];
		}
	} else if (i >= 0 && i < MAXFILES)
	{
		f = prg->fhandle[i];
	}
	if (f < 0)
	{
		GOTOS;
	}
	if (wflag == 0)
	{
		fstat(f, &s);
		SM_UW(t, time2dos(s.st_mtime));
		SM_UW(d, time2dos(s.st_mtime));
		DREG(0) = 0;
	} else if (wflag == 1)
	{
		/* TODO: get filename for utime() */
		DREG(0) = 0;
	} else
	{
		DREG(0) = 0;
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Fpipe(UW *handles)
{
	G(("Fpipe($%lx) ", VADDR(handles)));
	NYI;
}

/**************************************************************/

static int Fcntl(W fd, UL arg, UW cmd)
{
	G(("Fcntl(%d,$%lx,%u) ", fd, arg, cmd));
	NYI;
}

/**************************************************************/

static int Finstat(W fd)
{
	G(("Finstat(%d) ", fd));
	NYI;
}

/**************************************************************/

static int Foutstat(W fd)
{
	G(("Foutstat(%d) ", fd));
	NYI;
}

/**************************************************************/

static int Fgetchar(W fd, W mode)
{
	G(("Fgetchar(%d,%d) ", fd, mode));
	NYI;
}

/**************************************************************/

static int Fputchar(W fd, W mode)
{
	G(("Fputchar(%d,%d) ", fd, mode));
	NYI;
}

/**************************************************************/

static int Fselect(W timeout, UL rfds, UL wfds, UL dum)
{
	G(("Fselect(%d,$%lx,$%lx,$%lx ", timeout, rfds, wfds, dum));
	NYI;
}

/**************************************************************/

static int Fmidipipe(W pid, W in, W out)
{
	G(("Fmidipipe(%d,%d,%d) ", pid, in, out));
	NYI;
}

/**************************************************************/

static int Fxattr(W flag, char *pname, char *buf)
{
	char upath[PATH_MAX];
	char fullpath[PATH_MAX];
	struct stat s;
	
	G(("Fxattr(%d,%s,$%lx) ", flag, pname, VADDR(buf)));
	if (!st2unixpath(pname, fullpath, upath))
	{
		GOTOS;
	}
	if ((*(flag ? lstat : stat))(upath, &s) < 0)
	{
		DREG(0) = unix_to_tos_errno(errno);
		GRESULT;
		return GEMDOS_DONE;
	}
	stat_to_xattr(buf, &s);
	DREG(0) = GEMDOS_OK;
	GRESULT;
	return GEMDOS_DONE; 
}

/**************************************************************/

static int Flink(char *oldname, char *newname)
{
	char oldpath[PATH_MAX];
	char newpath[PATH_MAX];
	char fullpath[PATH_MAX];

	G(("Flink(%s,%s) ", oldname, newname));
	if (!st2unixpath(oldname, fullpath, oldpath) ||
		!st2unixpath(newname, fullpath, newpath))
	{
		GOTOS;
	}
	if (link(oldpath, newpath) < 0)
		DREG(0) = unix_to_tos_errno(errno);
	else
		DREG(0) = GEMDOS_OK;
	GRESULT;
	return GEMDOS_DONE;
}
 
/**************************************************************/

static int Fsymlink(char *oldname, char *newname)
{
	char oldpath[PATH_MAX];
	char newpath[PATH_MAX];
	char fullpath[PATH_MAX];

	G(("Fsymlink(%s,%s) ", oldname, newname));
	if (!st2unixpath(oldname, fullpath, oldpath) ||
		!st2unixpath(newname, fullpath, newpath))
	{
		GOTOS;
	}
	if (symlink(oldpath, newpath) < 0)
		DREG(0) = unix_to_tos_errno(errno);
	else
		DREG(0) = GEMDOS_OK;
	GRESULT;
	return GEMDOS_DONE;
}
 
/**************************************************************/

static int Freadlink(UW bufsize, char *buf, char *pname)
{
	char upath[PATH_MAX];
	char fullpath[PATH_MAX];
	
	G(("Freadlink(%d,$%lx,%s) ", bufsize, VADDR(buf), pname));
	if (!st2unixpath(pname, fullpath, upath))
	{
		GOTOS;
	}
	if (readlink(upath, buf, bufsize) < 0)
		DREG(0) = unix_to_tos_errno(errno);
	else
		DREG(0) = GEMDOS_OK;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Fchown(char *pname, W uid, W gid)
{
	char upath[PATH_MAX];
	char fullpath[PATH_MAX];
	
	G(("Fchown(%s,%d,%d) ", pname, uid, gid));
	if (!st2unixpath(pname, fullpath, upath))
	{
		GOTOS;
	}
	if (chown(upath, uid, gid) < 0)
		DREG(0) = unix_to_tos_errno(errno);
	else
		DREG(0) = GEMDOS_OK;
	return GEMDOS_DONE;
}
 
/**************************************************************/

static int Fchmod(char *pname, W mode)
{
	int unixmode;
	char upath[PATH_MAX];
	char fullpath[PATH_MAX];
		
	G(("Fchmod(%s,$%x) ", pname, mode));
	if (!st2unixpath(pname, fullpath, upath))
	{
		GOTOS;
	}
	unixmode = 0;
	if (mode & 0160000)
		unixmode = S_IFLNK;
	else if (mode & 0100000)
		unixmode = S_IFREG;
	else if (mode & 0060000)
		unixmode = S_IFBLK;
	else if (mode & 0040000)
		unixmode = S_IFDIR;
	else if (mode & 0020000)
		unixmode = S_IFCHR;
	else if (mode & 0120000)
		unixmode = S_IFIFO;
	if (mode & 04000)
		unixmode |= S_ISUID;
	if (mode & 02000)
		unixmode |= S_ISGID;
	if (mode & 01000)
		unixmode |= S_ISVTX;
	unixmode |= mode & 0777;
	if (chmod(upath, unixmode) < 0)
		DREG(0) = unix_to_tos_errno(errno);
	else
		DREG(0) = GEMDOS_OK;
	GRESULT;
	return GEMDOS_DONE;
}
 
/**************************************************************/
/**************************************************************/

static int Ptermres(UL keepcnt, W retcode)
{
	G(("Ptermres(%ld,%d), bp=$%lx ", keepcnt, retcode, get_bp()));
	GOTOS;
}

/**************************************************************/

static jmp_buf gemdos_jmp;

void gemdos_post(void)
{
	longjmp(gemdos_jmp, 1);
}

static void call_gemdos(int args_size)
{
	L save_pc = pc;
	
	if (setjmp(gemdos_jmp) == 0)
	{
		execute(CART_CALL_GEMDOS);
	}
	SP += args_size;
	pc = save_pc;
}

static int find_prog(char *file, char *upath, UL *size)
{
	DTA dta;
	
	if (!do_fsfirst(file, FA_RDONLY|FA_HIDDEN|FA_SYSTEM, &dta, upath))
		return FALSE;
	*size = LM_L((L *)dta.dta_size);
	return DREG(0) == 0;
}

static int init_prg(UL bp, UL parent_bp)
{
	PRG *prg;
	int i;
	
	if ((prg = malloc(sizeof(*prg))) == NULL)
		return FALSE;
		
	for (i = 0; i < MIN_FILE_HANDLE; i++)
	{
		prg->stdhandle[i] = LM_B(MEM(parent_bp+48+i));
	}
	for (i = 0; i < MAXFILES; i++)
		prg->fhandle[i] = INVALID_HANDLE;
	set_prg(bp, prg);
	prg->domain = DOM_TOS;
	prg->usrval = 0;
	return TRUE;
}

static void inherit_handles(UL child_pb, UL parent_bp)
{
	PRG *parent, *child;
	int fd;
	int i;
	
	parent = get_prg_bp(parent_bp);
	child = get_prg_bp(child_pb);
	if (parent && child)
	{
		for (i = 0; i < MIN_FILE_HANDLE; i++)
		{
			fd = child->stdhandle[i] - H_OFFSET;
			if (fd >= 0 && fd < MAXFILES)
			{
				child->fhandle[fd] = dup(parent->fhandle[fd]);
			}
		}
		child->usrval = parent->usrval;
	}
}
	
static int reloc(L bp, char *file, UL file_size)
{
	int f;
	char exec_header[28];
	L text_length, data_length, bss_length, sym_length, bss_start;
	UL prg_flags;
	UW abs_flag;
	UL offset;
	UB *offset_p;
	UB *reloc_p;
	UB rel;
	
	f = open(file, O_RDONLY);
	if (f < 0)
	{
		DREG(0) = GEMDOS_EFILNF;
		return FALSE;
	}
	if (read(f, exec_header, sizeof(exec_header)) != sizeof(exec_header))
	{
		DREG(0) = GEMDOS_EREADF;
		close(f);
		return FALSE;
	}
	if (LM_W((W *)exec_header) != 0x601a)
	{
		DREG(0) = GEMDOS_EPLFMT;
		close(f);
		return FALSE;
	}
	text_length = LM_L((L *)(exec_header+2));
	data_length = LM_L((L *)(exec_header+6));
	bss_length = LM_L((L *)(exec_header+10));
	sym_length = LM_L((L *)(exec_header+14));
	prg_flags = LM_UL((UL *)(exec_header+22));
	abs_flag = LM_W((W *)(exec_header+26));
	
	if ((text_length + data_length + sym_length + sizeof(exec_header)) > file_size)
	{
		DREG(0) = GEMDOS_EPLFMT;
		close(f);
		return FALSE;
	}
	if ((LM_L(MEM(bp+0)) + text_length + data_length + bss_length) > LM_L(MEM(bp+4)))
	{
		DREG(0) = GEMDOS_ENSMEM;
		close(f);
		return FALSE;
	}
	if (read(f, MEM(bp+256), (text_length+data_length)) < 0)
	{
		DREG(0) = GEMDOS_EREADF;
		close(f);
		return FALSE;
	}
	lseek(f, sym_length, 1);
	SM_L(MEM(bp+8), bp+256); /* text start */
	SM_L(MEM(bp+12), text_length);
	SM_L(MEM(bp+16), bp+256+text_length); /* data start */
	SM_L(MEM(bp+20), data_length);
	bss_start = bp+256+text_length+data_length;
	SM_L(MEM(bp+24), bss_start); /* bss start */
	SM_L(MEM(bp+28), bss_length);
	SM_L(MEM(bp+32), bp+128); /* p_dta */
	if (read(f, MEM(bss_start), file_size-text_length-data_length-sizeof(exec_header)) < 0)
	{
		DREG(0) = GEMDOS_EREADF;
		close(f);
		return FALSE;
	}
	close(f);
	if (abs_flag == 0)
	{
		offset = LM_L(MEM(bp+8));
		offset_p = (UB *)MEM(offset);
		reloc_p = (UB *)MEM(bss_start);
		if (LM_L((L *)(reloc_p)) != 0)
		{
			offset_p += LM_L((L *)(reloc_p));
			reloc_p += 4;
			SM_UL((UL *)offset_p, LM_UL((UL *)offset_p) + offset);
			for (;;)
			{
				rel = LM_UB(reloc_p);
				if (rel == 0)
					break;
				reloc_p++;
				if (rel == 1)
				{
					offset_p += 254;
				} else
				{
					offset_p += rel;
					SM_UL((UL *)offset_p, LM_UL((UL *)offset_p) + offset);
				}
			}
		}
	}
	bzero(MEM(bss_start), bss_length);
	return TRUE;
}

static int Pexec3(UL ptr1, UL ptr2, UL ptr3)
{
	char *file = MEM(ptr1);
	L bp, parent_bp;
	char upath[PATH_MAX];
	UL file_size;
	
	if (ptr1 == 0 || !find_prog(file, upath, &file_size))
	{
		DREG(0) = GEMDOS_EFILNF;
		GRESULT;
		return FALSE;
	}
	/*
	 * create basepage
	 */
	PUSH_UL(ptr3);
	PUSH_UL(ptr2);
	PUSH_UL(0);
	PUSH_UW(5);
	PUSH_UW(0x4b);
	call_gemdos(16);
	bp = DREG(0);
	if (bp < 0)
	{
		GRESULT;
		return FALSE;
	}

	parent_bp = get_bp();
	if (!init_prg(bp, parent_bp))
	{
		PUSH_UL(bp);
		PUSH_UW(0x49); /* Mfree(bp) */
		call_gemdos(6);
		DREG(0) = GEMDOS_ENSMEM;
		GRESULT;
		return FALSE;
	}
	/*
	 * load and relocate program image
	 */
	if (!reloc(bp, upath, file_size))
	{
		L err = DREG(0);
		
		PUSH_UL(bp);
		PUSH_UW(0x49); /* Mfree(bp) */
		call_gemdos(6);
		DREG(0) = err;
		GRESULT;
		return FALSE;
	}
	inherit_handles(bp, parent_bp);
	
	DREG(0) = bp;
	G(("bp=$%lx\n", bp));
	
	return TRUE;
}

static int Pexec(char *args)
{
	UW mode = LM_UW(args);
	UL ptr1 = LM_UL(args+2);
	UL ptr2 = LM_UL(args+6);
	UL ptr3 = LM_UL(args+10);
	UL parent;
	
	switch (mode)
	{
	case 0:
		/* load & go */
		G(("Pexec(%d,%s,%s,%s) by bp $%lx", mode,
			ptr1 ? (const char *)MEM(ptr1) : "NULL",
			ptr2 ? (const char *)MEM(ptr2) : "NULL",
			ptr3 ? (const char *)MEM(ptr3) : "NULL",
			get_bp()));
		if (Pexec3(ptr1, ptr2, ptr3))
		{
			UL bp = DREG(0);
			UW sversion;
			
			PUSH_UW(0x30);
			call_gemdos(2); /* Sversion() */
			sversion = LDREG_W(0);
			SM_L(args+2, 0);
			SM_L(args+6, bp);
			SM_L(args+10, 0);
			SM_W(args, 4);
			if (sversion >= 0x1500)
				SM_W(args, 6);
			return CALL_GEMDOS;
		}
		return GEMDOS_DONE;
	case 3:
		/* load, don't go */
		G(("Pexec(%d,%s,%s,%s) by bp $%lx ", mode,
			ptr1 ? (const char *)MEM(ptr1) : "NULL",
			ptr2 ? (const char *)MEM(ptr2) : "NULL",
			ptr3 ? (const char *)MEM(ptr3) : "NULL",
			get_bp()));
		Pexec3(ptr1, ptr2, ptr3);
		return GEMDOS_DONE;
	case 4:
		/* just go */
		G(("Pexec(%d,$%lx,$%lx,$%lx) by bp $%lx, start address $%lx ", mode, ptr1, ptr2, ptr3, get_bp(), LM_UL(MEM(ptr2+8)) ));
		GOTOS;
	case 5:
		/* create basepage */
		G(("Pexec(%d,%s,%s,%s) by bp $%lx ",
			mode,
			ptr1 ? (const char *)MEM(ptr1) : "NULL",
			ptr2 ? (const char *)MEM(ptr2) : "NULL",
			ptr3 ? (const char *)MEM(ptr3) : "NULL",
			get_bp()));
		PUSH_UL(ptr3);
		PUSH_UL(ptr1);
		PUSH_UL(ptr1);
		PUSH_UW(mode);
		PUSH_UW(0x4b);
		call_gemdos(16);
		if (DREG(0) >= 0)
		{
			parent = get_bp();
			init_prg(DREG(0), parent);
			inherit_handles(DREG(0), parent);
			G((" = $%lx\n", DREG(0)));
		} else
		{
			GRESULT;
		}
 		return GEMDOS_DONE;
	case 6:
		/* just go, then free */
		G(("Pexec(%d,$%lx,$%lx,$%lx) by bp $%lx, start address $%lx ", mode, ptr1, ptr2, ptr3, get_bp(), LM_UL(MEM(ptr2+8)) ));
		GOTOS;
	case 7:
		/* create basepage, respecting prgflags */
		G(("Pexec(%d,$%lx,%s,%s) by bp $%lx ",
			mode,
			ptr1,
			ptr2 ? (const char *)MEM(ptr2) : "NULL",
			ptr3 ? (const char *)MEM(ptr3) : "NULL",
			get_bp()));
		PUSH_UL(ptr3);
		PUSH_UL(ptr1);
		PUSH_UL(ptr1);
		PUSH_UW(mode);
		PUSH_UW(0x4b);
		call_gemdos(16);
		if (DREG(0) >= 0)
		{
			parent = get_bp();
			init_prg(DREG(0), parent);
			inherit_handles(DREG(0), parent);
			G((" = $%lx\n", DREG(0)));
		} else
		{
			GRESULT;
		}
 		return GEMDOS_DONE;
	case 100:
		/* MiNT: asynchron load & go */
		G(("Pexec(%d,%s,%s,%s) ",
			mode,
			ptr1 ? (const char *)MEM(ptr1) : "NULL",
			ptr2 ? (const char *)MEM(ptr2) : "NULL",
			ptr3 ? (const char *)MEM(ptr3) : "NULL"));
		GOTOS;
	case 101:
		/* MiNT: inherit handles */
		G(("Pexec(%d,$%lx,$%lx,$%lx) ", mode, ptr1, ptr2, ptr3));
		GOTOS;
	case 102:
		/* Terminate */
		G(("Pexec(%d,$%lx,$%lx,$%lx) ", mode, ptr1, ptr2, ptr3));
		GOTOS;
	case 104:
		/* MiNT: asynchron just go */
		G(("Pexec(%d,$%lx,$%lx,$%lx) ", mode, ptr1, ptr2, ptr3));
		GOTOS;
	case 106:
		/* MiNT: asynchron go, then free */
		G(("Pexec(%d,$%lx,$%lx,$%lx) ", mode, ptr1, ptr2, ptr3));
		GOTOS;
	case 107:
		/* MiNT: create basepage respecting prgflags */
		G(("Pexec(%d,$%lx,$%lx,$%lx) ", mode, ptr1, ptr2, ptr3));
		GOTOS;
	case 108:
		/* MagX: run accessory */
		G(("Pexec(%d,$%lx,$%lx,$%lx) ", mode, ptr1, ptr2, ptr3));
		GOTOS;
	case 200:
		/* MiNT: overlay */
		G(("Pexec(%d,$%lx,$%lx,$%lx) ", mode, ptr1, ptr2, ptr3));
		GOTOS;
	case 204:
		/* MiNT: overlay go */
		G(("Pexec(%d,$%lx,$%lx,$%lx) ", mode, ptr1, ptr2, ptr3));
		GOTOS;
	case 206:
		/* MiNT: overlay go and free */
		G(("Pexec(%d,$%lx,$%lx,$%lx) ", mode, ptr1, ptr2, ptr3));
		GOTOS;
	default:
		G(("Pexec(%d,$%lx,$%lx,$%lx) ", mode, ptr1, ptr2, ptr3));
		GOTOS;
	}
}

/**************************************************************/

static int do_pterm(int retcode)
{
	UL bp;
	PRG *prg;
	int i;
	
	bp = get_bp();
	prg = get_prg_bp(bp);
	if (prg != NULL)
	{
		for (i = 0; i < MAXFILES; i++)
			if (prg->fhandle[i] >= 0)
			{
				close(prg->fhandle[i]);
				check_close(prg, bp, i + H_OFFSET);
			}
		for (i = 0; i < MIN_FILE_HANDLE; i++)
			if (prg->stdhandle[i] >= H_OFFSET)
				close(prg->fhandle[prg->stdhandle[i]-H_OFFSET]);
		set_prg(get_bp(), NULL);
		free(prg);
	}
	GOTOS;
}

static int Pterm(W retcode)
{
	G(("Pterm(%d), bp=$%lx ", retcode, get_bp()));
	return do_pterm(retcode);
}

/**************************************************************/

static int Pterm0(void)
{
	G(("Pterm0(), bp=$%lx ", get_bp()));
	return do_pterm(0);
}

/**************************************************************/

static int Pwait(void)
{
	G(("Pwait() "));
	NYI;
}

/**************************************************************/

static int Pnice(W delta)
{
	G(("Pnice(%d) ", delta));
	NYI;
}

/**************************************************************/

static int Pgetpid(void)
{
	G(("Pgetpid() "));
	NYI;
}

/**************************************************************/

static int Pgetppid(void)
{
	G(("Pgetppid() "));
	NYI;
}

/**************************************************************/

static int Pgetpgrp(void)
{
	G(("Pgetpgrp() "));
	NYI;
}

/**************************************************************/

static int Psetpgrp(W pid, W pgrp)
{
	G(("Psetpgrp(%d,%d) ", pid, pgrp));
	NYI;
}

/**************************************************************/

static int Pgetuid(void)
{
	G(("Pgetuid() "));
	DREG(0) = getuid();
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Psetuid(W uid)
{
	G(("Psetuid(%d) ", uid));
	DREG(0) = 0;
	if (setuid(uid) < 0)
		DREG(0) = GEMDOS_EACCDN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Pkill(W pid, W sig)
{
	G(("Pkill(%d,%d) ", pid, sig));
	NYI;
}

/**************************************************************/

static int Psignal(W sig, UL handler)
{
	G(("Psignal(%d,$%lx) ", sig, handler));
	NYI;
}

/**************************************************************/

static int Pvfork(void)
{
	G(("Pvfork() "));
	NYI;
}

/**************************************************************/

static int Pgetgid(void)
{
	G(("Pgetgid() "));
	DREG(0) = getgid();
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Psetgid(W gid)
{
	G(("Psetgid(%d) ", gid));
	DREG(0) = 0;
	if (setgid(gid) < 0)
		DREG(0) =  GEMDOS_EACCDN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Psigblock(UL mask)
{
	G(("Psigblock($%lx) ", mask));
	NYI;
}

/**************************************************************/

static int Psigsetmask(UL mask)
{
	G(("Psigsetmask($%lx) ", mask));
	NYI;
}

/**************************************************************/

static int Pusrval(UL val)
{
	PRG *prg;
	
	G(("Pusrval($%lx) ", val));
	if ((prg = get_prg()) != NULL)
	{
		DREG(0) = prg->usrval;
		if (val != -1)
			prg->usrval = val;
		GRESULT;
		return GEMDOS_DONE;
	}
	GOTOS;
}

/**************************************************************/

static int Pdomain(W dom)
{
	PRG *prg;
	
	G(("Pdomain(%d) ", dom));
	prg = get_prg();
	if (prg)
	{
		DREG(0) = prg->domain;
		if (dom >= 0)
			prg->domain = dom;
		GRESULT;
		return GEMDOS_DONE;
	}
	GOTOS;
}

/**************************************************************/

static int Psigreturn(void)
{
	G(("Psigreturn() "));
	NYI;
}

/**************************************************************/

static int Pfork(void)
{
	G(("Pfork() "));
	NYI;
}

/**************************************************************/

static int Pwait3(W flags, UL rusage)
{
	G(("Pwait3(%d,$%lx) ", flags, rusage));
	NYI;
}

/**************************************************************/

static int Prusage(UL rusage)
{
	G(("Prusage($%lx) ", rusage));
	NYI;
}

/**************************************************************/

static int Psetlimit(W lim, L value)
{
	G(("Psetlimit(%d,%ld) ", lim, value));
	NYI;
}

/**************************************************************/

static int Pause(void)
{
	G(("Pause()" ));
	NYI;
}

/**************************************************************/

static int Psigpending(void)
{
	G(("Psigpending()" ));
	NYI;
}

/**************************************************************/

static int Pmsg(W mode, UL mbox, UL msg)
{
	G(("Pmsg(%d,%ld,$%lx) ", mode, mbox, msg));
	NYI;
}

/**************************************************************/

static int Prenice(W pid, W delta)
{
	G(("Prenice(%d,%d) ", pid, delta));
	NYI;
}

/**************************************************************/

static int Pumask(W mask)
{
	G(("Pumask(%x) ", mask));
	NYI;
}

/**************************************************************/

static int Psemaphore(W mode, L id, L timeout)
{
	G(("Psemaphore(%d,%ld,%ld) ", mode, id, timeout));
	NYI;
}

/**************************************************************/

static int Psigpause(UL mask)
{
	G(("Psigpause($%lx) ", mask));
	NYI;
}

/**************************************************************/

static int Psigaction(W sig, UL act, UL oact)
{
	G(("Psigaction(%d,$%lx,$%lx) ", sig, act, oact));
	NYI;
}

/**************************************************************/

static int Pgeteuid(void)
{
	G(("Pgeteuid() "));
	DREG(0) = geteuid();
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Pgetegid(void)
{
	G(("Pgeteuid() "));
	DREG(0) = getegid();
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Pwaitpid(W pid, W flag, UL rusage)
{
	G(("Pwaitpid(%d,%d,$%lx) ", pid, flag, rusage));
	NYI;
}

/**************************************************************/

static int Psigintr(W vec, W sig)
{
	G(("Psigintr(%d,%d) ", vec, sig));
	NYI;
}

/**************************************************************/

static int Pseteuid(W euid)
{
	G(("Pseteuid(%d) ", euid));
	DREG(0) = 0;
	if (seteuid(euid) < 0)
		DREG(0) = GEMDOS_EACCDN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Psetegid(W egid)
{
	G(("Psetegid(%d) ", egid));
	DREG(0) = 0;
	if (setegid(egid) < 0)
		DREG(0) = GEMDOS_EACCDN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Pgetauid(void)
{
	G(("Pgetauid() "));
	DREG(0) = GEMDOS_EINVFN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Psetauid(W auid)
{
	G(("Psetauid(%d) ", auid));
	DREG(0) = GEMDOS_EINVFN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Pgetgroups(W numgroups, W *groups)
{
	gid_t *list = NULL;
	int i, n;
	
	G(("Pgetgroups(%d,$%lx) ", numgroups, VADDR(groups)));
	if (numgroups > 0)
	{
		list = malloc(numgroups * sizeof(gid_t));
		if (!list)
		{
			DREG(0) = GEMDOS_ENSMEM;
			GRESULT;
			return GEMDOS_DONE;
		}
	}
	n = getgroups(numgroups, list);
	if (n < 0)
	{
		DREG(0) = GEMDOS_ERANGE;
		free(list);
		GRESULT;
		return GEMDOS_DONE;
	}
	for (i = 0; i < n; i++)
	{
		SM_W(groups+i, list[i]);
	}
	free(list);
	DREG(0) = n;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Psetgroups(W numgroups, W *groups)
{
	gid_t *list = NULL;
	int i, n;
	
	G(("Psetgroups(%d,$%lx) ", numgroups, VADDR(groups)));
	if (numgroups > 0)
	{
		list = malloc(numgroups * sizeof(gid_t));
		if (!list)
		{
			DREG(0) = GEMDOS_ENSMEM;
			GRESULT;
			return GEMDOS_DONE;
		}
	}
	for (i = 0; i < numgroups; i++)
		list[i] = LM_W(groups+i);
	n = setgroups(numgroups, list);
	if (n < 0)
	{
		if (errno == EPERM)
			DREG(0) = GEMDOS_EACCDN;
		else
			DREG(0) = GEMDOS_ERANGE;
		free(list);
		GRESULT;
		return GEMDOS_DONE;
	}
	DREG(0) = 0;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Psetreuid(W ruid, W euid)
{
	G(("Psetreuid(%d,%d) ", ruid, euid));
	DREG(0) = 0;
	if (setreuid(ruid, euid) < 0)
		DREG(0) = GEMDOS_EACCDN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Psetregid(W rgid, W egid)
{
	G(("Psetregid(%d,%d) ", rgid, egid));
	DREG(0) = 0;
	if (setregid(rgid, egid) < 0)
		DREG(0) = GEMDOS_EACCDN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/
/**************************************************************/

static int Maddalt(UL addr, UL size)
{
	G(("Maddalt($%lx,%ld) ", addr, size));
	GOTOS;
}

/**************************************************************/

static int Mxalloc(UL amount, UW mode)
{
	G(("Mxalloc(%ld,$%x) by BP $%lx ", amount, mode, get_bp()));
	GOTOS;
}

/**************************************************************/

static int Malloc(UL amount)
{
	G(("Malloc(%ld) by BP $%lx ", amount, get_bp()));
	GOTOS;
}

/**************************************************************/

static int Mfree(UL addr)
{
	G(("Mfree($%lx) by BP $%lx ", addr, get_bp()));
	GOTOS;
}

/**************************************************************/

static int Mshrink(UL addr, UL amount)
{
	G(("Mshrink($%lx,%ld) by BP $%lx ", addr, amount, get_bp()));
	GOTOS;
}

/**************************************************************/
/**************************************************************/

static int Sversion(void)
{
	G(("Sversion() "));
	GOTOS;
}

/**************************************************************/

static int Super(UL new_stack)
{
	G(("Super($%lx) ", new_stack));
	GOTOS;
}

/**************************************************************/

static int Sconfig(W mode, UL flags)
{
	G(("Sconfig(%d,$%lx) ", mode, flags));
	DREG(0) = GEMDOS_EINVFN;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Syield(void)
{
	G(("Syield() "));
	sleep(0);
	DREG(0) = 0;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Sysconf(W n)
{
	G(("Sysconf(%d) ", n));
	switch (n)
	{
	case -1:
		DREG(0) = 4;
		break;
	case 0:
		DREG(0) = 0x7fffffff;
		break;
	case 1:
		DREG(0) = 126;
		break;
	case 2:
		DREG(0) = MAXFILES;
		break;
	case 3:
		DREG(0) = 8;
		break;
	case 4:
		DREG(0) = 0x7fffffff;
		break;
	default:
		DREG(0) = GEMDOS_EINVFN;
		break;
	}
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Salert(char *msg)
{
	G(("Salert(%s) ", msg));
	fputs(msg, stderr);
	DREG(0) = 0;
	GRESULT;
	return GEMDOS_DONE;
}

/**************************************************************/

static int Suptime(UL *uptime, UL loadaverage[3])
{
	G(("Suptime($%lx,$%lx) ", VADDR(uptime), VADDR(loadaverage)));
	NYI;
}

/**************************************************************/

static int Scookie(W action, UL cookie)
{
	G(("Scookie(%d,$%lx) ", action, cookie));
	/* obsolete */
	GOTOS;
}

/**************************************************************/
/**************************************************************/

static int Tgetdate(void)
{
	G(("Tgetdate() "));
	GOTOS;
}

/**************************************************************/

static int Tsetdate(UW date)
{
	G(("Tsetdate($%x) ", date));
	GOTOS;
}

/**************************************************************/

static int Tgettime(void)
{
	G(("Tgettime() "));
	GOTOS;
}

/**************************************************************/

static int Tsettime(UW time)
{
	G(("Tsettime($%x)", time));
	GOTOS;
}

/**************************************************************/

static int do_alarm(L millisecs)
{
	NYI;
}

static int Talarm(L secs)
{
	G(("Talarm(%ld) ", secs));
	if (do_alarm(secs * 1000))
	{
		DREG(0) = (DREG(0) + 999) / 1000;
		return TRUE;
	}
	return FALSE;
}

/**************************************************************/

static int Tmalarm(L millisecs)
{
	G(("Tmalarm(%ld) ", millisecs));
	return do_alarm(millisecs);
}

/**************************************************************/

static int Tsetitimer(W which, UL interval, UL value, UL ointerval, UL ovalue)
{
	G(("Tsetimer(%d,$%lx,$%lx,$%lx,$%lx) ", which, interval, value, ointerval, ovalue));
	NYI;
}

/**************************************************************/
/**************************************************************/

int Gemdos(UL as)
{
	W n;
	UW ssr;
	char *args;
	int gemdos_done;
	
	sr &= ~MASK_CC_V;
	ssr = LM_UW(MEM(as - 4));
	if (!(ssr & SR_S))
		args = MEM(USP);
	else {
		args = MEM(as + 2);
		if (LM_UW(MEM(0x59e)) != 0)
			args += 2;
	}
	n = LM_W(args);
	args += 2;
	switch (n)
	{
	case 0:
		gemdos_done = Pterm0();
		break;
	case 1:
		gemdos_done = Cconin();
		break;
	case 2:
		gemdos_done = Cconout(LM_UW(args));
		break;
	case 3:
		gemdos_done = Cauxin();
		break;
	case 4:
		gemdos_done = Cauxout(LM_UW(args));
		break;
	case 5:
		gemdos_done = Cprnout(LM_UW(args));
		break;
	case 6:
		gemdos_done = Crawio(LM_UW(args));
		break;
	case 7:
		gemdos_done = Crawcin();
		break;
	case 8:
		gemdos_done = Cnecin();
		break;
	case 9:
		gemdos_done = Cconws((char *) MEM(LM_UL(args)));
		break;
	case 10:
		gemdos_done = Cconrs((char *) MEM(LM_UL(args)));
		break;
	case 11:
		gemdos_done = Cconis();
		break;
	case 14:
		gemdos_done = Dsetdrv(LM_UW(args));
		break;
	case 16:
		gemdos_done = Cconos();
		break;
	case 17:
		gemdos_done = Cprnos();
		break;
	case 18:
		gemdos_done = Cauxis();
		break;
	case 19:
		gemdos_done = Cauxos();
		break;
	case 20:
		gemdos_done = Maddalt(LM_UL(args), LM_UL(args+4));
		break;
	case 25:
		gemdos_done = Dgetdrv();
		break;
	case 26:
		gemdos_done = Fsetdta(LM_UL(args));
		break;
	case 32:
		gemdos_done = Super(LM_UL(args));
		break;
	case 42:
		gemdos_done = Tgetdate();
		break;
	case 43:
		gemdos_done = Tsetdate(LM_UW(args));
		break;
	case 44:
		gemdos_done = Tgettime();
		break;
	case 45:
		gemdos_done = Tsettime(LM_UW(args));
		break;
	case 47:
		gemdos_done = Fgetdta();
		break;
	case 48:
		gemdos_done = Sversion();
		break;
	case 49:
		gemdos_done = Ptermres(LM_UL(args), LM_UW(args+4));
		break;
	case 51:
		gemdos_done = Sconfig(LM_W(args), LM_UL(args+2));
		break;
	case 54:
		gemdos_done = Dfree(MEM(LM_UL(args)), LM_UW(args+4));
		break;
	case 57:
		gemdos_done = Dcreate((char *) MEM(LM_UL(args)));
		break;
	case 58:
		gemdos_done = Ddelete((char *) MEM(LM_UL(args)));
		break;
	case 59:
		gemdos_done = Dsetpath((char *) MEM(LM_UL(args)));
		break;
	case 60:
		gemdos_done = Fcreate((char *) MEM(LM_UL(args)), LM_UW(args+4));
		break;
	case 61:
		gemdos_done = Fopen((char *) MEM(LM_UL(args)), LM_UW(args+4));
		break;
	case 62:
		gemdos_done = Fclose(LM_W(args));
		break;
	case 63:
		gemdos_done = Fread(LM_W(args), LM_UL(args+2), LM_UL(args+6));
		break;
	case 64:
		gemdos_done = Fwrite(LM_W(args), LM_UL(args+2), LM_UL(args+6));
		break;
	case 65:
		gemdos_done = Fdelete((char *) MEM(LM_UL(args)));
		break;
	case 66:
		gemdos_done = Fseek(LM_UL(args), LM_W(args+4), LM_UW(args+6));
		break;
	case 67:
		gemdos_done = Fattrib(MEM(LM_UL(args)), LM_UW(args+4), LM_UW(args+6));
		break;
	case 68:
		gemdos_done = Mxalloc(LM_UL(args), LM_UW(args+4));
		break;
	case 69:
		gemdos_done = Fdup(LM_W(args));
		break;
	case 70:
		gemdos_done = Fforce(LM_W(args), LM_W(args+2));
		break;
	case 71:
		gemdos_done = Dgetpath((char *) MEM(LM_UL(args)), LM_UW(args+4));
		break;
	case 72:
		gemdos_done = Malloc(LM_UL(args));
		break;
	case 73:
		gemdos_done = Mfree(LM_UL(args));
		break;
	case 74:
		gemdos_done = Mshrink(LM_UL(args+2), LM_UL(args+6));
		break;
	case 75:
		gemdos_done = Pexec(args);
		break;
	case 76:
		gemdos_done = Pterm(LM_UW(args));
		break;
	case 78:
		gemdos_done = Fsfirst((char *) MEM(LM_UL(args)), LM_UW(args+4));
		break;
	case 79:
		gemdos_done = Fsnext();
		break;
	case 86:
		gemdos_done = Frename(MEM(LM_UL(args+2)), MEM(LM_UL(args+6)));
		break;
	case 87:
		gemdos_done = Fdatime(MEM(LM_UL(args)), MEM(LM_UL(args) + 2), LM_W(args+4), LM_UW(args+6));
		break;
	
	/* MiNT calls */
	case 255:
		gemdos_done = Syield();
		break;
	case 256:
		gemdos_done = Fpipe((UW *)MEM(LM_UL(args)));
		break;
	case 260:
		gemdos_done = Fcntl(LM_W(args), LM_UL(args+2), LM_UW(args+6));
		break;
	case 261:
		gemdos_done = Finstat(LM_W(args));
		break;
	case 262:
		gemdos_done = Foutstat(LM_W(args));
		break;
	case 263:
		gemdos_done = Fgetchar(LM_W(args), LM_W(args+2));
		break;
	case 264:
		gemdos_done = Fputchar(LM_W(args), LM_W(args+2));
		break;
	case 265:
		gemdos_done = Pwait();
		break;
	case 266:
		gemdos_done = Pnice(LM_W(args));
		break;
	case 267:
		gemdos_done = Pgetpid();
		break;
	case 268:
		gemdos_done = Pgetppid();
		break;
	case 269:
		gemdos_done = Pgetpgrp();
		break;
	case 270:
		gemdos_done = Psetpgrp(LM_W(args), LM_W(args+2));
		break;
	case 271:
		gemdos_done = Pgetuid();
		break;
	case 272:
		gemdos_done = Psetuid(LM_W(args));
		break;
	case 273:
		gemdos_done = Pkill(LM_W(args), LM_W(args+2));
		break;
	case 274:
		gemdos_done = Psignal(LM_W(args), LM_UL(args+2));
		break;
	case 275:
		gemdos_done = Pvfork();
		break;
	case 276:
		gemdos_done = Pgetgid();
		break;
	case 277:
		gemdos_done = Psetgid(LM_W(args));
		break;
	case 278:
		gemdos_done = Psigblock(LM_UL(args));
		break;
	case 279:
		gemdos_done = Psigsetmask(LM_UL(args));
		break;
	case 280:
		gemdos_done = Pusrval(LM_UL(args));
		break;
	case 281:
		gemdos_done = Pdomain(LM_W(args));
		break;
	case 282:
		gemdos_done = Psigreturn();
		break;
	case 283:
		gemdos_done = Pfork();
		break;
	case 284:
		gemdos_done = Pwait3(LM_W(args), LM_UL(args+2));
		break;
	case 285:
		gemdos_done = Fselect(LM_W(args), LM_UL(args+2), LM_UL(args+6), LM_UL(args+10));
		break;
	case 286:
		gemdos_done = Prusage(LM_UL(args));
		break;
	case 287:
		gemdos_done = Psetlimit(LM_W(args), LM_L(args+2));
		break;
	case 288:
		gemdos_done = Talarm(LM_L(args));
		break;
	case 289:
		gemdos_done = Pause();
		break;
	case 290:
		gemdos_done = Sysconf(LM_W(args));
		break;
	case 291:
		gemdos_done = Psigpending();
		break;
	case 292:
		gemdos_done = Dpathconf((char *)MEM(LM_UL(args)), LM_W(args+4));
		break;
	case 293:
		gemdos_done = Pmsg(LM_W(args), LM_UL(args+2), LM_UL(args+6));
		break;
	case 294:
		gemdos_done = Fmidipipe(LM_W(args), LM_W(args+2), LM_W(args+4));
		break;
	case 295:
		gemdos_done = Prenice(LM_W(args), LM_W(args+2));
		break;
	case 296:
		gemdos_done = Dopendir((char *)MEM(LM_UL(args)), LM_W(args+4));
		break;
	case 297:
		gemdos_done = Dreaddir(LM_W(args), LM_L(args+2), (char *)MEM(LM_UL(args+6)));
		break;
	case 298:
		gemdos_done = Drewinddir(LM_L(args));
		break;
	case 299:
		gemdos_done = Dclosedir(LM_L(args));
		break;
	case 300:
		gemdos_done = Fxattr(LM_W(args), (char *)MEM(LM_UL(args+2)), MEM(LM_UL(args+6)));
		break;
	case 301:
		gemdos_done = Flink((char *)MEM(LM_UL(args)), (char *)MEM(LM_UL(args+4)));
		break;
	case 302:
		gemdos_done = Fsymlink((char *)MEM(LM_UL(args)), (char *)MEM(LM_UL(args+4)));
		break;
	case 303:
		gemdos_done = Freadlink(LM_W(args), (char *)MEM(LM_UL(args+2)), (char *)MEM(LM_UL(args+6)));
		break;
	case 304:
		gemdos_done = Dcntl(LM_W(args), (char *)MEM(LM_UL(args+2)), LM_UL(args+6));
		break;
	case 305:
		gemdos_done = Fchown((char *)MEM(LM_UL(args)), LM_W(args+4), LM_W(args+6));
		break;
	case 306:
		gemdos_done = Fchmod((char *)MEM(LM_UL(args)), LM_UW(args+4));
		break;
	case 307:
		gemdos_done = Pumask(LM_W(args));
		break;
	case 308:
		gemdos_done = Psemaphore(LM_W(args), LM_L(args+2), LM_L(args+6));
		break;
	case 309:
		gemdos_done = Dlock(LM_W(args), LM_W(args+2));
		break;
	case 310:
		gemdos_done = Psigpause(LM_UL(args));
		break;
	case 311:
		gemdos_done = Psigaction(LM_W(args), LM_UL(args+2), LM_UL(args+6));
		break;
	case 312:
		gemdos_done = Pgeteuid();
		break;
	case 313:
		gemdos_done = Pgetegid();
		break;
	case 314:
		gemdos_done = Pwaitpid(LM_W(args), LM_W(args+2), LM_UL(args+4));
		break;
	case 315:
		gemdos_done = Dgetcwd((char *)MEM(LM_UL(args)), LM_UW(args+4), LM_UW(args+6));
		break;
	case 316:
		gemdos_done = Salert((char *)MEM(LM_UL(args)));
		break;
	case 317:
		gemdos_done = Tmalarm(LM_L(args));
		break;
	case 318:
		gemdos_done = Psigintr(LM_W(args), LM_W(args+2));
		break;
	case 319:
		gemdos_done = Suptime((UL *)MEM(LM_UL(args)), (UL *)MEM(LM_UL(args+4)));
		break;
	/* 320 undefined */
	/* 321 undefined */
	case 322:
		gemdos_done = Dxreaddir(LM_W(args), LM_L(args+2), MEM(LM_UL(args+6)), LM_UL(args+10), LM_UL(args+14));
		break;
	case 323:
		gemdos_done = Pseteuid(LM_W(args));
		break;
	case 324:
		gemdos_done = Psetegid(LM_W(args));
		break;
	case 325:
		gemdos_done = Pgetauid();
		break;
	case 326:
		gemdos_done = Psetauid(LM_W(args));
		break;
	case 327:
		gemdos_done = Pgetgroups(LM_W(args), (W *)MEM(LM_UL(args+2)));
		break;
	case 328:
		gemdos_done = Psetgroups(LM_W(args), (W *)MEM(LM_UL(args+2)));
		break;
	case 329:
		gemdos_done =  Tsetitimer(LM_W(args), LM_UL(args+2), LM_UL(args+6), LM_UL(args+10), LM_UL(args+14));
		break;
	case 330:
		gemdos_done = Scookie(LM_W(args), LM_UL(args+2));
		break;
	/* 331 undefined */
	/* 332 undefined */
	/* 333 undefined */
	case 334:
		gemdos_done = Psetreuid(LM_W(args), LM_W(args+2));
		break;
	case 335:
		gemdos_done = Psetregid(LM_W(args), LM_W(args+2));
		break;
	/* 336 undefined */
	/* 337 undefined */
	case 338:
		gemdos_done = Dreadlabel((char *)MEM(LM_UL(args)), (char *)MEM(LM_UL(args+4)), LM_UW(args+8));
 		break;
 	case 339:
 		gemdos_done = Dwritelabel((char *)MEM(LM_UL(args)), (char *)MEM(LM_UL(args+4)));
 		break;
	default:
		G(("Gemdos %d\n", n));
		gemdos_done = FALSE;
		break;
	}
	SET_N();
	UNSET_Z();
	if (gemdos_done)
		SET_Z();
	return gemdos_done;
}

void init_gemdos(void)
{
#if UNIX_FS
	if (!gemdos_ok && gemdos_drives)
	{
		fprintf(stderr,"Initializing GEMDOS patches...\n");
		SM_L(MEM(CART_OLD_GEMDOS), LM_L(EXCEPTION_VECTOR(T_TRAP_GEMDOS)));
		SM_L(EXCEPTION_VECTOR(T_TRAP_GEMDOS), CART_NEW_GEMDOS);
		gemdos_ok = TRUE;
	}
#endif
#if 0
	drive_bits |= gemdos_drives;
#endif
}
